@@ -7,10 +7,20 @@ Author: Nikolay Lysenko
 
 import itertools
 import math
+import re
+import string
+from collections import Counter
+from functools import cache
 from typing import Any, Callable, Optional
 
 from .fragment import Event, Fragment
-from .music_theory import IntervalTypes, N_SEMITONES_PER_OCTAVE, get_type_of_interval
+from .music_theory import (
+    IntervalTypes,
+    N_SEMITONES_PER_OCTAVE,
+    get_mapping_from_pitch_class_to_diatonic_scales,
+    get_type_of_interval,
+)
+from .utils import compute_rolling_aggregate
 
 
 SCORING_SETS_REGISTRY_TYPE = dict[
@@ -19,6 +29,41 @@ SCORING_SETS_REGISTRY_TYPE = dict[
 ]
 
 
+def evaluate_absence_of_aimless_fluctuations(
+        fragment: Fragment, penalties: dict[int, float], window_size: int
+) -> float:
+    """
+    Evaluate melodic fluency based on absence of aimless fluctuations within narrow ranges.
+
+    :param fragment:
+        a fragment to be evaluated
+    :param penalties:
+        mapping from range (in semitones) covered within a window to penalty applicable to ranges
+        of not greater size; it is recommended to set maximum penalty to 1
+    :param window_size:
+        size of rolling window (in events)
+    :return:
+        multiplied by -1 penalty for narrowness averaged over all windows (including those that
+        are not penalized)
+    """
+    numerator = 0
+    denominator = 0
+    for melodic_line in fragment.melodic_lines:
+        pitches = [event.position_in_semitones for event in melodic_line]
+        pitches = [x for x in pitches if x is not None]
+        rolling_mins = compute_rolling_aggregate(pitches, min, window_size)[window_size-1:]
+        rolling_maxs = compute_rolling_aggregate(pitches, max, window_size)[window_size-1:]
+        borders = zip(rolling_mins, rolling_maxs)
+        for lower_border, upper_border in borders:
+            range_width = upper_border - lower_border
+            raw_penalties = [v for k, v in penalties.items() if k >= range_width]
+            penalty = max(raw_penalties) if raw_penalties else 0
+            numerator -= penalty
+            denominator += 1
+    score = numerator / denominator
+    return score
+
+
 def evaluate_absence_of_doubled_pitch_classes(fragment: Fragment) -> float:
     """
     Evaluate absence of vertical intervals that are whole multipliers of octave.
@@ -430,6 +475,224 @@ def evaluate_harmony_dynamic(
     return score
 
 
+def evaluate_local_diatonicity(
+        fragment: Fragment, depth: int = 2, scale_types: Optional[tuple[str]] = None
+) -> float:
+    """
+    Evaluate presence of diatonic scales at short periods of time.
+
+    If for every short interval there is a diatonic scale containing all its pitches,
+    the fragment might be called pantonal with frequent modulations between diatonic scales.
+
+    :param fragment:
+        a fragment to be evaluated
+    :param depth:
+        duration of a time period as number of successive sonorities
+    :param scale_types:
+        types of diatonic scales to be tested; this tuple may include the following values:
+        'major', 'natural_minor', 'harmonic_minor', 'dorian', 'phrygian', 'lydian', 'mixolydian',
+        'locrian', and 'whole_tone'; however, keep in mind that due to relative modes there is no
+        need to include all these values; if this argument is not passed, its value is set to
+        `('major', 'harmonic_minor', 'whole_tone')` which covers all supported scales
+    :return:
+        minus one multiplied by average fraction of pitches that are out of the most fitting
+        to the current short interval diatonic scale (with averaging over all short intervals)
+    """
+    score = 0
+    scale_types = scale_types or ('major', 'harmonic_minor', 'whole_tone')
+    pitch_class_to_diatonic_scales = get_mapping_from_pitch_class_to_diatonic_scales(scale_types)
+    nested_pitch_classes = [[]]
+    for i in range(depth - 1):
+        nested_pitch_classes.append([event.pitch_class for event in fragment.sonorities[i]])
+    for i in range(depth - 1, len(fragment.sonorities)):
+        nested_pitch_classes.pop(0)
+        nested_pitch_classes.append([event.pitch_class for event in fragment.sonorities[i]])
+        pitch_classes = [x for y in nested_pitch_classes for x in y if x != 'pause']
+        counter = Counter()
+        for pitch_class in pitch_classes:
+            counter.update(pitch_class_to_diatonic_scales[pitch_class])
+        n_pitch_classes_from_best_scale = counter.most_common(1)[0][1]
+        score -= 1 - n_pitch_classes_from_best_scale / len(pitch_classes)
+    n_periods = len(fragment.sonorities) - depth + 1
+    score /= n_periods
+    return score
+
+
+@cache
+def get_mapping_from_interval_size_to_character() -> dict[Optional[int], str]:
+    """
+    Get mapping from interval size (in semitones) to its single-char encoding.
+
+    :return:
+        mapping from size (in semitones) of a directed non-compound interval to
+        its single-char representation; `None` as key relates to pauses
+    """
+    intervals = list(range(-N_SEMITONES_PER_OCTAVE, N_SEMITONES_PER_OCTAVE + 1)) + [None]
+    result = {interval: letter for interval, letter in zip(intervals, string.ascii_lowercase)}
+    return result
+
+
+def encode_interval(interval: Optional[int]) -> str:
+    """
+    Encode an interval with a single character.
+
+    :param interval:
+        interval size (in semitones)
+    :return:
+        single-character encoding of the interval
+    """
+    mapping = get_mapping_from_interval_size_to_character()
+    if interval is not None:
+        interval = math.copysign(abs(interval) % N_SEMITONES_PER_OCTAVE, interval)
+    return mapping[interval]
+
+
+def encode_line_intervals(melodic_lines: list[list[Event]]) -> list[str]:
+    """
+    Encode intervals from all melodic lines
+
+    :param melodic_lines:
+        melodic lines
+    :return:
+        strings with encoded intervals for each melodic line
+    """
+    encoded_lines = []
+    for melodic_line in melodic_lines:
+        encoded_line = ''
+        for previous_event, next_event in zip(melodic_line, melodic_line[1:]):
+            if previous_event.pitch_class == 'pause' or next_event.pitch_class == 'pause':
+                interval = None
+            else:
+                interval = next_event.position_in_semitones - previous_event.position_in_semitones
+            encoded_line += encode_interval(interval)
+        encoded_lines.append(encoded_line)
+    return encoded_lines
+
+
+def generate_elision_patterns(
+        motif: tuple[int], original_pattern: str, run: bool = False
+) -> list[str]:
+    """
+    Generate search patterns for modifications of the original motif with one pitch omitted.
+
+    :param motif:
+        sequence of directed interval sizes (in semitones) between successive pitches of a motif
+    :param original_pattern:
+        search pattern (in encoded lines) corresponding to the motif
+    :param run:
+        flag whether to generate any patterns
+    :return:
+        generated patterns corresponding to elisions of the motif
+    """
+    patterns = []
+    if not run:
+        return patterns
+    not_first = f'[^{original_pattern[0]}]'
+    first_omission_pattern = not_first + original_pattern[1:]
+    patterns.append(first_omission_pattern)
+    for i in range(len(motif) - 1):
+        motif_with_omission = motif[:i] + (motif[i] + motif[i + 1],) + motif[i + 2:]
+        omission_pattern = ''.join([encode_interval(x) for x in motif_with_omission])
+        patterns.append(omission_pattern)
+    not_last = f'[^{original_pattern[-1]}]'
+    last_omission_pattern = original_pattern[:-1] + not_last
+    patterns.append(last_omission_pattern)
+    return patterns
+
+
+@cache
+def generate_regexps_for_intervallic_motif(
+        motif: tuple[int],
+        inversion: bool = True, reversion: bool = True,
+        elision: bool = False, inverted_elision: bool = False, reverted_elision: bool = False
+) -> list['_sre.SRE_Pattern']:
+    """
+    Generate regular expressions that match intervallic motif and its modifications.
+
+    :param motif:
+        sequence of directed interval sizes (in semitones) between successive pitches of a motif
+    :param inversion:
+        flag whether to include inversion of the original motif
+    :param reversion:
+        flag whether to include reversion of the original motif
+    :param elision:
+        flag whether to include modifications of the original motif with one pitch omitted
+    :param inverted_elision:
+        flag whether to include modifications of the inverted motif with one pitch omitted;
+        if `inversion` is set to `False`, this flag affects nothing
+    :param reverted_elision:
+        flag whether to include modifications of the reverted motif with one pitch omitted;
+        if `reversion` is set to `False`, this flag affects nothing
+    :return:
+        compiled regular expressions
+    """
+    original_pattern = ''.join([encode_interval(x) for x in motif])
+    patterns = [original_pattern]
+    patterns.extend(generate_elision_patterns(motif, original_pattern, elision))
+    if inversion:
+        inverted_motif = tuple(-x for x in motif)
+        inverted_pattern = ''.join([encode_interval(x) for x in inverted_motif])
+        patterns.append(inverted_pattern)
+        patterns.extend(
+            generate_elision_patterns(inverted_motif, inverted_pattern, inverted_elision)
+        )
+    if reversion:
+        reverted_motif = motif[::-1]
+        reverted_pattern = original_pattern[::-1]
+        patterns.append(reverted_pattern)
+        patterns.extend(
+            generate_elision_patterns(reverted_motif, reverted_pattern, reverted_elision)
+        )
+    patterns = list(set(patterns))  # In particular, duplicates may occur due to symmetries.
+    regexps = [re.compile(pattern) for pattern in patterns]
+    return regexps
+
+
+def evaluate_presence_of_intervallic_motif(
+        fragment: Fragment, motif: list[int], min_n_occurrences: list[int],
+        inversion: bool = True, reversion: bool = True,
+        elision: bool = False, inverted_elision: bool = False, reverted_elision: bool = False
+) -> float:
+    """
+    Evaluate presence of used-defined intervallic motif.
+
+    :param fragment:
+        a fragment to be evaluated
+    :param motif:
+        sequence of directed interval sizes (in semitones) between successive pitches of a motif
+    :param min_n_occurrences:
+        minimum numbers of motif occurrences for each melodic line
+    :param inversion:
+        flag whether to include inversion of the original motif
+    :param reversion:
+        flag whether to include reversion of the original motif
+    :param elision:
+        flag whether to include modifications of the original motif with one pitch omitted
+    :param inverted_elision:
+        flag whether to include modifications of the inverted motif with one pitch omitted;
+        if `inversion` is set to `False`, this flag affects nothing
+    :param reverted_elision:
+        flag whether to include modifications of the reverted motif with one pitch omitted;
+        if `reversion` is set to `False`, this flag affects nothing
+    :return:
+        minus one multiplied by relative lack of motif occurrences
+    """
+    regexps = generate_regexps_for_intervallic_motif(
+        tuple(motif), inversion, reversion, elision, inverted_elision, reverted_elision
+    )
+    encoded_lines = encode_line_intervals(fragment.melodic_lines)
+    numerator = 0
+    denominator = 0
+    for threshold, encoded_line in zip(min_n_occurrences, encoded_lines):
+        n_occurrences = 0
+        for regexp in regexps:
+            n_occurrences += len(re.findall(regexp, encoded_line))
+        numerator -= max(threshold - n_occurrences, 0)
+        denominator += threshold
+    score = numerator / denominator
+    return score
+
+
 def evaluate_rhythmic_homogeneity(fragment: Fragment) -> float:
     """
     Evaluate rhythmic homogeneity between all measures except the last one.
@@ -522,6 +785,7 @@ def get_scoring_functions_registry() -> dict[str, Callable]:
         registry of scoring functions
     """
     registry = {
+        'absence_of_aimless_fluctuations': evaluate_absence_of_aimless_fluctuations,
         'absence_of_doubled_pitch_classes': evaluate_absence_of_doubled_pitch_classes,
         'absence_of_simultaneous_skips': evaluate_absence_of_simultaneous_skips,
         'absence_of_voice_crossing': evaluate_absence_of_voice_crossing,
@@ -530,6 +794,8 @@ def get_scoring_functions_registry() -> dict[str, Callable]:
         'consistency_of_rhythm_with_meter': evaluate_consistency_of_rhythm_with_meter,
         'dissonances_preparation_and_resolution': evaluate_dissonances_preparation_and_resolution,
         'harmony_dynamic': evaluate_harmony_dynamic,
+        'intervallic_motif': evaluate_presence_of_intervallic_motif,
+        'local_diatonicity': evaluate_local_diatonicity,
         'rhythmic_homogeneity': evaluate_rhythmic_homogeneity,
         'smoothness_of_voice_leading': evaluate_smoothness_of_voice_leading,
         'stackability': evaluate_stackability,
